#!/usr/bin/perl -w
# Author: Alex Efros <powerman@asdfGroup.com>
# Desc -- Simultaneous downloading of several URLs from single process,
#     without threads usage
# Idea -- Use non-blocking sockets for not waiting for anything (connecting,
#     reading, etc.).
#     Use POWER::Multi::DNS for performing DNS query in background.
#     Use asynchronous algorithm for processing urls - each url contain
#     information about it current stage (parsing, dns-querying, connecting,
#     writing http query, reading reply) and all urls processing in loop,
#     executing one stage at time for every url.
#
package POWER::Multi::GET;
use Socket;
use Fcntl;
use Errno qw( EINPROGRESS EAGAIN );
use POWER::lib qw( time sleep );
use POWER::NB::Resolver 1.11;
use POWER::NB::IO 1.11;		    # for ungzip large files
use Time::Local;		    # for cookie's expires
use IPC::Open3;			    # for gzip without tmp files
BEGIN { eval "use Compress::Zlib" } # instead of IPC::Open3 and POWER::NB::IO
use strict;
use vars qw($VERSION $Max_Traffic $Max_URL %Timeout %url $Proxy $SLEEP
    %CacheDNS $Max_CacheDNS $Max_Redirect $Headers $Headers_gzip $Use_gzip);

$VERSION	= "2.03";
$Max_Traffic    = 10240000;	# maximum of read bytes in second
$Max_URL	= 100;		# maximum of simultaneous processing url
%CacheDNS	= ();
$Max_CacheDNS	= 1000;		# %DNS maximum size
$Max_Redirect	= 5;
$SLEEP          = 0.01;
$Headers	= "User-Agent: Mozilla/4.0 (compatible; MSIE 5.0; Windows 98; DigExt)\r\n";
$Headers_gzip	= "Accept-Encoding: gzip, deflate\r\n";
$Use_gzip	= undef;
%Timeout        = (		# timeout in seconds:
        Resolving   => 0.01,	# DNS query in progress
        Connecting  => 0.01,	# connect in progress
	Writing	    => 0.01,	# written part of HTTP query
        Reading     => 0.01,	# not ready data to read
        Connect     => 30,	# how much wait for connect
        Read        => 30,	# how much wait after last successful read
    );
$Proxy	    = undef;		# HTTP proxy configuration
my $Now     = time();		# current second
my $Max_Buf = 102400;		# buffer size for sysread()
my $Traffic = 0;		# count of downloaded bytes in current second
my $TrafficTime	= 0;
my @Func    = (\&_parse,\&_connect,\&_connecting,\&_write,\&_read,
    \&_ungzip,\&_cookie,\&_redirect,\&_finish);
my %NextFunc= ();		# hash to easy find next function
@NextFunc{@Func} = @Func[1..$#Func];    
%url        = ();		# current url
my $Resolver= tie my %DNS, 'POWER::NB::Resolver';

{ my %month; @month{qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec)} = (0 .. 11);
sub add_cookie {
    my (%cookie) = @_;
    if ($cookie{text}) {
	return if $cookie{text} !~ /^\s*(?:([^=]*)=)?([^;]*)(.*)/mi;
	(@cookie{'name','value'}, my $params) = ($1, $2, $3);
	$params =~ /.*;\s*$_(?:\s*=\s*([^;]*?))?\s*(?:;|$)/i and
	    $cookie{$_}=$1 for qw(domain path expires secure);
    }
    $cookie{$_} ||= "", $cookie{$_} =~ s/^\s*|\s*$//g for qw(name value);
    $cookie{domain} = $url{_host} unless $cookie{domain};
    ($cookie{path}) = $url{_path} =~ m!(.*/)! unless defined $cookie{path};
    return if $cookie{domain} !~ /^[^\s;]+$/;
    return if $cookie{path}   !~ /^[^\s;]*$/;
    $cookie{domain} =~ s/^\.//;
    if (not $cookie{trust}++) {
	return if $url{_host} !~ /(?:^|\.)\Q$cookie{domain}\E$/;
	return if $url{_path} !~ /^\Q$cookie{path}\E/;
	return if $cookie{domain} ne $url{_host} and 
	    @{[ $cookie{domain}=~/(?<=.)\.(?=.)/g ]} <
	    ( @{[ $url{_host}=~/(?<=.)\.(?=.)/g ]} <= 2 ? 1 : 2 );
    }
    if ($cookie{expires}) {
	return unless my ($day,$mon,$year,$hour,$min,$sec) = $cookie{expires} =~ /^(?:Mon|Tue|Wed|Thu|Fri|Sat|Sun), (\d\d)[- ](Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[- ](\d?\d?\d\d) ([01]\d|2[0-3]):([0-5]\d):([0-5]\d)(?:\s*$| GMT| UTC)/i;
	$year = $year > 50 ? "19$year" : "20$year" if $year < 100;
	$cookie{expires_time} = eval 'local $SIG{__DIE__};timegm($sec,$min,$hour,$day,$month{$mon},$year)';
	return if (gmtime($cookie{expires_time}))[4] != $month{$mon};
    }
    return if $cookie{expires_time} !~ /^\d*$/;
    # TODO      secure=something ?
    my $cid = join(";", @cookie{'name','domain','path'},exists $cookie{secure});
    $cookie{add_time} ||= $url{_cookie_hash}{$cid}{add_time} || timelocal(gmtime());
    $url{_cookie_hash}{$cid} = \%cookie;
}}

# parse url, normalize url, setup HTTP-query text
sub _parse {
    @url{'_sheme','_user','_pass','_host','_port','_path','_params'} = 
	$url{url} =~ m{ 
	^(?:([a-zA-Z]+)://)?	    # http://
	(?:([^:@/]+)		    # user
	(?::([^@/]+))?		    # :pass
	@)?			    # @
				    # host name as defined in RFC 1738
	# ? TODO {_host}     ???
	((?:(?:[a-zA-Z0-9]-*)*[a-zA-Z0-9]\.)*(?:(?:[a-zA-Z0-9]-*)*[a-zA-Z0-9]))
	(?::(\d*))?		    # :port
	(?:(\/[^?]*)(?:\?(.*))?)?$  # /path?params
	}x or return("Bad url syntax",0);
    $url{_sheme}   = lc($url{_sheme} || 'http');
    $url{_host}    = lc $url{_host};
    my $port = getservbyname($url{_sheme}, 'tcp');
    $url{_port}  ||= $port || getservbyname('http', 'tcp');
    $url{_path}  ||= '/';
    $url{_url}     = $url{_sheme}."://".$url{_host};
    $url{_url}    .= ":".$url{_port} if $url{_port} != $port;
    $url{_url}    .= $url{_path};
    add_cookie(trust=>1, %$_) for @{$url{cookie}||[]};
    for (values %{$url{_cookie_hash}}) {
	next if $url{_host} !~ /(?:^|\.)\Q$_->{domain}\E$/;
	next if $url{_path} !~ /^\Q$_->{path}\E/;
	next if timelocal(gmtime()) > $_->{expires_time};
    }
    $url{_cookie_header} = [ 
	map  { ($_->{name}?$_->{name}."=":"").$_->{value} }
	grep { not exists $_->{secure} or $url{_sheme} eq "https" }
	grep { $url{_host} =~ /(?:^|\.)\Q$_->{domain}\E$/ }
	grep { $url{_path} =~ /^\Q$_->{path}\E/ }
	grep { ! $_->{expires_time} or timelocal(gmtime()) <= $_->{expires_time} }
	sort { $a->{add_time} <=> $b->{add_time} }
	values %{$url{_cookie_hash}} ];
    $url{_cookie_header} = @{$url{_cookie_header}} ?
	"Cookie: ".join("; ",@{$url{_cookie_header}})."\r\n" : "";
    my ($phost,$pport) = ($url{proxy}||$Proxy)=~m/^(?:http:\/\/)?([^:\/]+):(\d+)\/?$/;
    return("bad proxy",0) if (($url{proxy}||$Proxy) and not $phost);
    $url{_resolve} = $phost || $url{ip} || $url{_host};
    $url{_iport}   = $pport || $url{_port};
    $url{_request} = $url{_msg} = 
	( defined($url{post_data})	? "POST "		: "GET "      ).
	( $phost			? $url{_url}		: $url{_path} ).
	( length($url{_params})		? "?$url{_params}"	: ""	      ).
	" HTTP/1.0\r\nHost: ".$url{_host}."\r\n".$Headers.$url{_cookie_header}.
	( $Use_gzip			? $Headers_gzip		: ""	      ).
        ( defined($url{add_to_header})	? $url{add_to_header}	: ""	      ).
	( defined($url{post_data})	?
	    "Content-type: application/x-www-form-urlencoded\r\n".
	    "Content-length: ".length($url{post_data})."\r\n\r\n".
	    $url{post_data}					: "\r\n"      );
}
# get IP, ask for connect
sub _connect {
    return("DNS answer not ready",$Timeout{Resolving},\&_connect)
	until $CacheDNS{$url{_resolve}} || $DNS{$url{_resolve}};
    $CacheDNS{$url{_resolve}} ||= delete $DNS{$url{_resolve}};
    return($CacheDNS{$url{_resolve}}{ERR},0) unless $CacheDNS{$url{_resolve}}{IP};
    $url{_iaddr} = pack("C4", split (/\./, $CacheDNS{$url{_resolve}}{IP}));
    %CacheDNS = () if $Max_CacheDNS < keys %CacheDNS;
    { local *SOCK; $url{_SOCK} = *SOCK }
    select( (select($url{_SOCK}), $|=1)[0] );
    socket($url{_SOCK}, AF_INET, SOCK_STREAM, getprotobyname('tcp')) or
    return("socket: $!",0);
    fcntl($url{_SOCK}, F_SETFL, O_NONBLOCK) or return("fcntl: $!",0);
    connect($url{_SOCK}, sockaddr_in($url{_iport}, $url{_iaddr})) or
        ($! != EINPROGRESS and return("connect: $!",0));
    $url{_CONNECT_timeout} = $Now + $Timeout{Connect};
}
# wait for connect
sub _connecting {
    return("CONNECT timeout",0) if $Now >= $url{_CONNECT_timeout};
    vec(my $w="",fileno($url{_SOCK}),1)=1;
    return("wait for connect",$Timeout{Connecting},\&_connecting) unless 
    select(undef,$w,undef,0);
    return("connect: $!",0) if $!=ord getsockopt($url{_SOCK},SOL_SOCKET,SO_ERROR);
}
# write HTTP query to socket
sub _write {
    my $n = syswrite($url{_SOCK},$url{_msg});
    return("wait for write",$Timeout{Writing},\&_write) if $! == EAGAIN;
    return("write: $!",0) if $! or not defined $n;
    return("written part of message",$Timeout{Writing},\&_write) if 
        length( $url{_msg}=substr($url{_msg},$n) );
    $url{_READ_timeout} = $Now + $Timeout{Read};
}
# check $Max_Traffic, $Timeout{Read}, read HTTP header and content
sub _read {
    return("traffic limit",1,\&_read) if $Traffic >= $Max_Traffic;
    my $n = sysread($url{_SOCK}, my $buf, Min($Max_Buf,$Max_Traffic-$Traffic));
    if ($! == EAGAIN) {
    return("READ timeout",0) if $Now >= $url{_READ_timeout};
    return("data not ready to reading",$Timeout{Reading},\&_read);
    }
    return("read: $!",0) if $! or not defined $n;
    $url{_READ_timeout} = $Now + $Timeout{Read};
    $Traffic += $n;
    $url{_content} .= substr($buf,0,$n);
    # cut HTTP-header when we read it
    $url{_content} =~ s/^(.*?\r?\n)\r?\n/$url{_header}=$1;""/se unless $url{_header};
    return("read in progress",0,\&_read) if $n != 0;
}
# check $Use_gzip, ungzip content if needed
{ my ($gzip) = grep { -f $_ && -x $_ } qw( /usr/bin/gzip /bin/gzip );
sub _ungzip {
    return unless $Use_gzip and $url{_header} =~ /^Content-Encoding: gzip\r?$/mi;
    # TODO  :
    # 1)   Compress::Zlib     
    # 2)    gzip  :  while(1)   stage
    # 3)    gzip    _read
    # 4)  memGunzip  _read     gzread  tied FD
    # 5)    (fifo?) +  gzip  ?
    #  2,3     POWER::Multi::GET  
    #     -  gzip-.
    #  3,4   ungzip   callback {readed},
    # ..  {readed}       
    #   /bin/gzip  gzread(),    _read().
    if (exists &Compress::Zlib::memGunzip) {
	my $out = Compress::Zlib::memGunzip(my $in = $url{_content});
	return("ungzip failed",0) unless $out;
	$url{_content} = $out;
    } else {
	my ($pid, $w, $r, $e, $deflate); local ($/, $SIG{PIPE}, $SIG{CHLD});
	$pid = open3($w, $r, $e, $gzip, '-dc', '-');
	{ my ($nb_w, $nb_r) = (nb_file($w), nb_file($r));
	  nb_write($nb_w, $url{_content});
	  while (1) {
	    close($w), undef($nb_w) if $nb_w and not $nb_w->{OUT};
	    return($nb_w->{ERR},0)  if $nb_w and not nb_io($nb_w, skip_read  => 1);
	    return($nb_r->{ERR},0)  if           not nb_io($nb_r, skip_write => 1);
	    last if $nb_r->{EOF};
	    $deflate .= $nb_r->{IN};
	    $nb_r->{IN} = "";
	  }
	}
	waitpid $pid, 0;	    #   !!!
	return($deflate,0) if $deflate =~ /^open3:/ or $deflate =~ s/^\ngzip:/gzip:/;
	$url{_content} = $deflate;
    }
}}
# check header and remember correct cookies
sub _cookie {
    add_cookie(trust=>0, text=>$1) while $url{_header} =~ /^Set-Cookie:[ \t]*(.*?)\r?$/mig;
    $url{_cookie} = [ values %{$url{_cookie_hash}} ];
}
# check Max_Redirect, process auto-redirects with cookie support
sub _redirect {
    # TODO    <META HTTP-EQUIV=Refresh ...>
    return if $url{_header} !~ /^Location:\s+(\S+)\r?$/mi;
    return("too many redirects",0) if $url{_redirect} >= $Max_Redirect;
    exists $url{$_} and $url{"_orig_$_"} ||= $url{$_} for qw(url ip post_data);
    $url{url} = abs_url($1, $url{_url});
    $url{cookie} = $url{_cookie};
    $url{_redirect}++;
    delete @url{"ip", "post_data", grep {m/^_(?!redirect$|orig_)/} keys %url};
    return("redirected",0,$Func[0]);
}
# restore original url, ip and post_data after auto-redirects
sub _finish {
    exists $url{"_orig_$_"} and $url{$_}=$url{"_orig_$_"} for qw(url ip post_data);
}
# download all urls, check $Max_URL
{ my $download_timeout;
sub Download (\%;&$) {
    my $urls=0;
    local $Use_gzip = defined($Use_gzip) ? $Use_gzip : not(defined($_[1]));
    foreach ( (grep {   defined $_->{_state}} values %{$_[0]}),
	      (grep {not exists $_->{_state}} values %{$_[0]}) ) {
	@{$_}{"_ok","_err","_state"} = ("ERR", "Download timed out"), next
	    if defined $_[2] and time() > ($download_timeout ||= time()+$_[2]);
	last if ++$urls >= $Max_URL;
	local *url = $_;
	$Now = time();
	$TrafficTime = time(), $Traffic = 0 if time() - $TrafficTime >= 1;
	next if defined $url{_timeout} and $url{_timeout} > $Now;
	my ($err,$timeout,$newstate) = &{ $url{_state} ||= $Func[0] };
	@url{'_state','_err','_timeout'} = ! defined $timeout ?
	    ($NextFunc{$url{_state}},"") : ($newstate,$err,$Now+$timeout);
    } continue {
	if (not defined $_->{_state}) {
	    $_->{_ok} = $_->{_err} ? "ERR" : "OK";
	    $_->{downloaded}->($_) if $_->{downloaded};
	}
    }
    &{$_[1]} if $_[1];          # call user-defined function
    sleep( $SLEEP );            # lower CPU usage
    goto &Download if $urls or grep {not exists $_->{_state} or
	defined $_->{_state}} values %{$_[0]};
    undef $download_timeout;
}}
1;
